from typing import Protocol, Any, NamedTuple
from collections.abc import Iterable

import numpy as np
import numpy.typing as npt

Int = int | np.int_
Num = Int | float | np.float_


class Transformer(Protocol):
    def fit(self, *args, **kwargs) -> None:
        ...

    def transform(self, *args, **kwargs) -> Any:
        ...

    def fit_transform(self, *args, **kwargs) -> Any:
        ...


class LabelTransformer(Transformer, Protocol):
    """LabelTransformer"""

    def fit(self, time: Iterable[Num], event: Iterable[Int]) -> None:
        ...

    def transform(
        self,
        time: Iterable[Num],
        event: Iterable[Int],
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        ...

    def fit_transform(
        self,
        time: Iterable[Num],
        event: Iterable[Int],
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        ...

    @property
    def max_time(self) -> Num:
        ...

    @property
    def cuts(self) -> npt.NDArray[np.floating]:
        ...


class SurvData(NamedTuple):
    features: npt.NDArray[np.float_]
    event_time_disc: npt.NDArray[np.int_]
    event_status_disc: npt.NDArray[np.int_]
    event_time_cont: npt.NDArray[np.float_]
    event_status_cont: npt.NDArray[np.int_]



import logging

from pytest import fixture
from pytest import mark
from lightning import Trainer

from discotime.datasets import Mgus2, DataConfig
from discotime.models import LitSurvModule, ModelConfig

# configure logging at the root level of Lightning
logging.getLogger("lightning").setLevel(logging.ERROR)


@fixture
def mgus2_dm():
    dm = Mgus2(
        DataConfig(
            batch_size=60, n_time_bins=20, discretization_scheme="number"
        )
    )
    dm.prepare_data()
    return dm


def test_data_setup(mgus2_dm: Mgus2):
    mgus2_dm.setup(stage="fit")
    assert mgus2_dm.time_range == (2.0, 373.0)

    # hparams should be accessible
    assert set(mgus2_dm.hparams.keys()) == {"data_config", "seed"}

    assert len(mgus2_dm.dset_train[0]) == 5
    assert len(mgus2_dm.dset_test[0]) == 5

    assert len(next(iter(mgus2_dm.train_dataloader()))) == 5
    assert len(next(iter(mgus2_dm.test_dataloader()))) == 5

    # number of cuts is n_time_bins + 1
    assert len(mgus2_dm.cuts) == (mgus2_dm.hparams.data_config.n_time_bins + 1)  # type: ignore


@mark.filterwarnings("ignore")
def test_model_integration(mgus2_dm: Mgus2):
    model = LitSurvModule(ModelConfig())
    trainer = Trainer(
        accelerator="cpu",
        devices=1,
        max_epochs=20,
        logger=False,
        enable_checkpointing=False,
    )
    trainer.fit(model, mgus2_dm)
    trainer.test(model, mgus2_dm)

    assert True



from typing import Any, Optional, TypeVar
import torch

TensorLike = TypeVar("TensorLike")
"Object that can be converted to a tensor"


class Interpolate2D:
    """Perform stepwise linear interpolation of a discrete function."""

    @torch.no_grad()
    def __init__(
        self, xs: TensorLike, ys: TensorLike, dim: int = -1, oob: str = "error"
    ) -> None:
        """Fit piecewise linear model.

        Args:
            xs (`Tensor`): an 1D tensor of real values
            ys (`Tensor`): an ND tensor of real values. The length of `ys` on
                the interpolation axis must have the same length as `xs`.
            dim: along what axis should the interpolation be done?
            oob: how should values in `x` to __call__ be handled if they're
                outside the range of values in `xs` seen during fitting?
                Possible values to `oob` are "error", "extrapolate" or "clip".
        """

        if oob not in (allowed := {"error", "clip", "extrapolate"}):
            raise ValueError(f"{oob=} is not in {allowed}")

        xs = torch.as_tensor(xs).squeeze()
        ys = torch.as_tensor(ys)

        assert xs.device == ys.device

        if xs.ndim != 1:
            raise ValueError("`xs` must be a 1D tensor.")

        # put interpolation axis last
        ys = torch.swapaxes(ys, -1, dim)

        # obtain slopes and intercepts
        self.m = torch.diff(ys) / torch.diff(xs)
        self.b = ys[..., :-1] - (self.m * xs[:-1])

        self.xs = xs
        self.xrange = (torch.min(self.xs), torch.max(self.xs))
        self.dim = dim
        self.oob = oob

    @torch.no_grad()
    def __call__(self, xs: TensorLike) -> torch.Tensor:
        xs = torch.as_tensor(xs).squeeze()
        xs = torch.atleast_1d(xs)

        if xs.ndim != 1:
            raise ValueError("`xs` must be convertable to a 1D tensor.")

        if (is_oob := (xs < self.xrange[0]) | (xs > self.xrange[1])).any():
            if self.oob == "error":
                raise ValueError(f"{xs[is_oob]} is outside the range  `xs`.")

        if self.oob == "clip":
            xs = torch.clamp(xs, *self.xrange)

        # search self.xs to find what slope/intercept to use
        idx = torch.searchsorted(self.xs, xs) - 1
        idx[idx == -1] = 0

        if self.oob == "extrapolate":
            idx = torch.clamp(idx, min=0, max=self.m.shape[-1] - 1)

        # calculate ys and swap interpolation axis back
        ys = self.m[..., idx] * xs + self.b[..., idx]
        ys = ys.swapaxes(self.dim, -1)

        return ys


class KaplanMeier:
    """Simple implementation of the Kaplan-Meier estimator.

    Args:
        time: event times.
        event: event indicator (0/1) where 0 is censoring.

    Example:
        >>> km = KaplanMeier(time=[0, 1.5, 1.3, 3], event=[0, 1, 0, 0])
        >>> km(0)
        tensor([1.])
        >>> km([0, 1.0, 1.1, 1.5, 5])
        tensor([1.0000, 1.0000, 1.0000, 0.5000, 0.5000])
    """

    @torch.no_grad()
    def __init__(
        self, time: TensorLike, event: TensorLike, lag: int = 0
    ) -> None:
        _time = torch.as_tensor(time).squeeze()
        _event = torch.as_tensor(event).squeeze()

        if _event.ndim > 1 or _time.ndim > 1:
            raise ValueError("`time` and `event` should be 1D tensors")

        if _event.unique().shape > (2,):
            raise ValueError("there are more than 2 unique values in `event`")

        idx = torch.argsort(_time)
        _time, _event = _time[idx], _event[idx]

        # observed event times, which is guaranteed to include zero
        tj = torch.cat((torch.tensor([0]), _time[_event == 1])).unique()

        # count failure times
        tjm = torch.searchsorted(tj, _time[_event == 1], side="right")
        mj = torch.bincount(tjm, minlength=len(tj) + 1)[1:]

        # count censoring times
        tjq = torch.searchsorted(tj, _time[_event == 0], side="right")
        qj = torch.bincount(tjq, minlength=len(tj) + 1)[1:]

        nj = torch.roll(len(_time) - torch.cumsum(mj + qj, 0), 1)
        nj[0] = len(_time)

        # product limit formula
        self._sj = torch.cumprod((nj - mj) / nj, 0)
        self._tj, self._nj, self._mj = tj, nj, mj
        self.lag = lag

    @torch.no_grad()
    def __call__(self, time: TensorLike) -> torch.Tensor:
        """Obtain Kaplan-Meier estimates for each timepoint.

        Args:
            time (num | seq[num]): `t`'s for which `km(t)` will be returned.
        """

        _time = torch.as_tensor(time).squeeze()

        if _time.ndim > 1:
            raise ValueError("`time` is not a 1D tensor")

        idx = torch.searchsorted(self._tj, _time, side="right") - 1 - self.lag
        idx = torch.clamp(idx, 0, len(self._sj))
        return torch.atleast_1d(self._sj[idx])

    @torch.no_grad()
    def percentile(self, p: TensorLike) -> torch.Tensor:
        """Obtain approximate timepoint t such that P(t) = p.

        The stepwise Kaplan-Meier estimator is piecewise linearly interpolated
        such that unique timepoints can be obtained.
        """
        p = torch.atleast_1d(torch.as_tensor(p).squeeze())

        if not torch.all((0 <= p) & (p <= 1)):
            raise ValueError(
                "`p` is a probability and all values should be between 0 and 1"
            )

        return Interpolate2D(1 - self._sj, self._tj, oob="clip")(1 - p)


class AalenJohansen:
    """Obtain cumulative incidence curves with the Aalen-Johansen method.

    Args:
        time: event times
        event: event indicator (0/1/../c) with 0=censored
    """

    @torch.no_grad()
    def __init__(
        self,
        time: TensorLike,
        event: TensorLike,
        n_causes: Optional[int] = None,
    ) -> None:
        _time = torch.as_tensor(time).squeeze()
        _event = torch.as_tensor(event).squeeze().int()

        if _event.ndim > 1 or _time.ndim > 1:
            raise ValueError("`time` and `event` should be 1D tensors")

        if _event.unique().shape < (2,):
            raise ValueError("there are less than 2 unique values in `event`")

        idx = torch.argsort(_time)
        _time, _event = _time[idx], _event[idx]

        # observed event times, which is guaranteed to include zero
        tj = torch.cat((torch.tensor([0]).to(_time), _time)).unique()

        # nj is the number at risk at each of the intervals in tj
        qj = torch.bincount(torch.searchsorted(tj, _time) + 1)[:-1]
        nj = len(_time) - torch.cumsum(qj, dim=0)

        # mj is the number lost to any cause at each interval
        mj = torch.bincount(
            torch.searchsorted(tj, _time[_event != 0]), minlength=len(tj)
        )

        # lagged survival which starts with P(0) = 1
        sj = torch.cumprod((nj - mj) / nj, dim=0).roll(1)
        sj[0] = 1

        n_risks = n_causes if n_causes else torch.max(_event).item()
        assert isinstance(n_risks, int)  # type narrowing

        # cause-specific incidences
        ci = torch.zeros((len(tj), n_risks), dtype=torch.double)
        for e in range(1, n_risks + 1):
            te = _time[_event == e]
            mcj = torch.bincount(torch.searchsorted(tj, te), minlength=len(tj))
            ci[:, e - 1] = torch.cumsum(sj * (mcj / nj), dim=0)

        self._tj = tj
        self._sj = sj
        self._ci = ci

    @torch.no_grad()
    def __call__(self, time: TensorLike) -> torch.Tensor:
        """Obtain Kaplan-Meier estimates for each timepoint.

        Args:
            time (num | seq[num]): `t`'s for which `km(t)` will be returned.
        """

        _time = torch.as_tensor(time).squeeze().to(self._tj)

        if _time.ndim > 1:
            raise ValueError("`time` is not a 1D tensor")

        if torch.any(_time < 0):
            raise ValueError("can not understand negative values in `time`")

        idx = torch.searchsorted(self._tj, _time, side="right") - 1
        return torch.atleast_1d(self._ci[idx])


class IPCW:
    @torch.no_grad()
    def __init__(self, time=TensorLike, event=TensorLike) -> None:
        t = torch.as_tensor(time).squeeze()
        e = torch.as_tensor(event).squeeze()

        if e.ndim > 1 or e.ndim > 1:
            raise ValueError("`time` and `event` should be 1D tensors")

        idx = torch.argsort(t)
        t, e = t[idx], e[idx]

        # observed times
        tj = torch.unique(torch.cat((torch.tensor([0]), t)))

        mj = torch.bincount(
            torch.searchsorted(tj, t[e == 0]), minlength=len(tj)
        )
        rj = torch.bincount(
            torch.searchsorted(tj, t[e != 0]), minlength=len(tj)
        )
        qj = rj + mj.roll(1).index_put((torch.tensor(0),), torch.tensor(0))

        # at risk at each interval
        nj = t.shape[0] - torch.cumsum(qj, dim=0)

        self._tj = tj
        self._mj = mj
        self._hj = mj / nj
        self._sj = torch.cumprod(1 - (mj / nj), dim=0)

    @torch.no_grad()
    def __call__(self, time: TensorLike, lag: int = 0) -> torch.Tensor:
        t = torch.as_tensor(time).squeeze()

        if t.ndim > 1:
            raise ValueError("`time` is not a 1D tensor")

        if torch.any(t < 0):
            raise ValueError("negative values in `time`")

        idx = torch.searchsorted(self._tj, t, side="right") - 1 - lag
        idx = torch.clamp(idx, 0)
        return torch.atleast_1d(self._sj[idx])



import warnings
import numpy as np
import pytest

from discotime.datasets.utils import LabelDiscretizer
from discotime.datasets.utils import (
    _discretize_labels,
    _cuts_number,
)


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        ((0.1, 0), (0, 0)),
        ((0.6, 0), (1, 0)),
        ((0.6, 1), (0, 1)),
        ((1.6, 2), (1, 2)),
        ((2.1, 0), (2, 0)),
        ((3.1, 1), (2, 0)),
        ((3.8, 2), (2, 0)),
    ],
)
def test_label_discretization(input, expected):
    cuts = np.array([0, 1, 2, 3])
    assert _discretize_labels(input[0], input[1], cuts=cuts) == expected


def test_discretization_scheme_number():
    time = [1, 2, 3, 4, 5]
    event = [1, 1, 1, 1, 1]
    expected = [0, 1, 2, 3, 4, 5]
    cuts = _cuts_number(n_bins=5, max_time=5, time=time, event=event)
    assert cuts == pytest.approx(expected)


class TestLabTransDiscrete:
    def test_interval_discretization(self, survival_data_2):
        time, event = survival_data_2
        ltd = LabelDiscretizer("interval", 10)
        dt, de = ltd.fit_transform(time, event)
        assert len(set(dt)) == 10
        assert len(set(np.round(np.diff(ltd.cuts), 4))) == 1
        assert len(set(map(lambda x: len(x), (time, event, dt, de)))) == 1

    def test_quantile_discretization(self, survival_data_2):
        time, event = survival_data_2
        ltd = LabelDiscretizer("number", 10)
        dt, de = ltd.fit_transform(time, event)
        assert len(set(dt)) == 10
        assert len(set(np.diff(ltd.cuts))) != 1
        assert len(set(map(lambda x: len(x), (time, event, dt, de)))) == 1

    def test_manual_discretization(self, survival_data_2):
        time, event = survival_data_2
        ltd = LabelDiscretizer(cut_points=[0, 5, 10, 15])
        with pytest.warns(UserWarning, match="ignoring fit()"):
            dt, de = ltd.fit_transform(time, event)
        assert len(set(dt)) == 3
        assert len(set(map(lambda x: len(x), (time, event, dt, de)))) == 1



from typing import TypeVar
from itertools import pairwise

import torch
from discotime.utils import AalenJohansen, IPCW

TensorLike = TypeVar("TensorLike")
"Object that can be converted to a tensor"


class BrierScore:
    """Brier score of survival model with right-censored data.

    Args:
        survival_train: sequence of `t` timepoints to evaluate.
    """

    @torch.no_grad()
    def __init__(self, survival_train: tuple[TensorLike, TensorLike]) -> None:
        futime, status = survival_train
        futime = torch.as_tensor(futime).squeeze()
        status = torch.as_tensor(status).squeeze()

        if futime.ndim > 1 or status.ndim > 1:
            raise ValueError("`futime` and `status` should be 1D tensors")

        self.ipcw = IPCW(futime, status)

    @torch.no_grad()
    def __call__(
        self,
        estimates: TensorLike,
        timepoints: TensorLike,
        survival_test: tuple[TensorLike, TensorLike],
    ) -> torch.Tensor:
        """Calculate Brier score of survival model.

        Args:
            estimates: an array with shape (`m`, `t`, `e`), where `m` is the
                batch size, `t` is the number of time bins, and `e` is number
                of competing causes/risks.
            timepoints: array of timepoints with size `t` at which the
                values in `estimates` are obtained.
            survival_test: tuple of observed time/event data.
        """

        St = torch.as_tensor(estimates)
        n_obs, n_time, n_risks = St.shape

        tau = torch.as_tensor(timepoints)
        if len(tau) != n_time:
            raise ValueError(
                "size of `timepoints` do not match dim 1 of estimates."
            )

        futime, status = survival_test
        futime = torch.as_tensor(futime).squeeze()
        status = torch.as_tensor(status).squeeze()

        if futime.ndim > 1 or status.ndim > 1:
            raise ValueError("`futime` and `status` should be 1D tensors")

        if len(futime) != n_obs:
            raise ValueError("length of `futime` is not equal to batch size.")

        if len(status) != n_obs:
            raise ValueError("length of `status` is not equal to batch size.")

        # has individual experienced event of interest yet?
        I1 = torch.logical_and(
            futime.view(-1, 1, 1) <= tau.view(1, -1, 1),
            status.view(-1, 1, 1) == torch.arange(1, n_risks + 1),
        )
        # ... did something else happen?
        # fmt: off
        I2 = (
            (futime.view(-1, 1, 1) <= tau.view(1, -1, 1))
            .logical_and(status.view(-1, 1, 1) != torch.zeros_like(I1))
            .logical_and(status.view(-1, 1, 1) != torch.arange(1, n_risks + 1))
        )
        # fmt: on
        # ... or is individual still in the risk set?
        I3 = futime.view(-1, 1, 1) > tau.view(1, -1, 1)

        # The IPCW is clamped such that one individual is not "dominating" the
        # score. see Kvamme & Borgan. "The Brier Score under Administrative
        # Censoring: Problems and a Solution."
        Gt = torch.clamp(self.ipcw(tau, lag=0).reshape(1, -1, 1), min=1e-4)
        GTi = torch.clamp(self.ipcw(futime, lag=1).reshape(-1, 1, 1), min=1e-4)

        results = I1 * (1 - St) ** 2 / GTi
        results += I2 * (St**2) / GTi
        results += I3 * (St**2) / Gt

        return torch.mean(results, dim=0)


class BrierScoreScaled:
    """Index of prediction accuracy (IPA) for right-censored survival models.

    In the context of survival analysis, the IPA is also known as the Brier
    skill score (BSS).
    """

    @torch.no_grad()
    def __init__(
        self,
        survival_train: tuple[TensorLike, TensorLike],
        eval_grid: TensorLike,
        integrate: bool = False,
    ) -> None:
        """
        Args:
            timepoints: sequence of `t` timepoints to evaluate.
            causes: causes for which the brier score is calculated.
            integrate: should the time-dependent IPA be integrated?
                (default: True)
        """
        eval_grid = torch.as_tensor(eval_grid)

        if len({*eval_grid}) != len(eval_grid):
            raise ValueError("values in `timepoints` are not unique")

        if not all(a <= b for a, b in pairwise(eval_grid)):
            raise ValueError("values in `timepoints` are not sorted")

        if integrate and len(eval_grid) == 1:
            raise ValueError(
                "at least two values in `timepoints` are required when "
                "self.integrate = True"
            )

        self.brier_score = BrierScore(survival_train=survival_train)
        self.eval_grid = eval_grid
        self.integrate = integrate

    @torch.no_grad()
    def __call__(
        self,
        estimates: TensorLike,
        survival_test: tuple[TensorLike, TensorLike],
    ) -> torch.Tensor:
        """Calculate the index of prediction accuracy.

        Args:
            estimates: array[batch, timepoints, cause]
                model estimates of cause probabilities.
            time: observed follow-up.
            event: event indicator at time with 0 as censoring.

        If there are multiple timepoints to consider, then the IPA is
        numerically integrated using the trapezoidal method.
        """
        phi_test = torch.as_tensor(estimates)
        n_causes = phi_test.shape[-1]
        tau = self.eval_grid

        futime, status = survival_test
        futime = torch.as_tensor(futime).squeeze()
        status = torch.as_tensor(status).squeeze()

        cic = AalenJohansen(futime, status, n_causes)(tau)
        phi_null = cic.unsqueeze(0).expand_as(phi_test)

        bs_test = self.brier_score(phi_test, tau, survival_test)
        bs_null = self.brier_score(phi_null, tau, survival_test)

        if self.integrate:
            bs_null = torch.trapezoid(bs_null, tau.view(-1, 1), dim=0)
            bs_test = torch.trapezoid(bs_test, tau.view(-1, 1), dim=0)

        # avoid zero division
        bs_test[bs_null == 0] += 1e-10
        bs_null[bs_null == 0] += 1e-10

        return 1 - bs_test / bs_null



from itertools import pairwise, count, islice

import torch
import pytest
from hypothesis import given, example
from hypothesis import strategies as st

from discotime.metrics import BrierScore, BrierScoreScaled
from discotime.utils import AalenJohansen, IPCW


@given(
    tau=st.lists(st.floats(0, 0.699), min_size=1, max_size=100),
)
def test_brier_score_1(tau, survival_data_2):
    """Before any event occurs (t=0.7) the following is true:

    - An estimate of 100% gives a brier score of 1
    - An estimate of 0% gives a brier score of 0
    - An estimate of 50% gives a brier score of 0.25

    """
    time, event = survival_data_2
    brier_score = BrierScore(survival_train=(time, event))

    all_ones = torch.ones((len(time), len(tau), 2))
    all_zeros = torch.zeros_like(all_ones)
    all_half = torch.full_like(all_ones, 0.5)

    assert torch.all(brier_score(all_ones, tau, (time, event)) == 1)
    assert torch.all(brier_score(all_zeros, tau, (time, event)) == 0)
    assert torch.all(brier_score(all_half, tau, (time, event)) == 0.25)


def brier_score_nested(estimates, timepoints, survival_test):
    """Implementation of the Brier score using nested for loops.

    Relatively inefficient, but the code is easier to read than the one relying
    on torch broadcasting rules. This implementation should give the same
    result as the vectorized one, otherwise a bug/error have been introduced
    somewhere - most likely in the "efficient" version.
    """
    futime, status = map(torch.as_tensor, survival_test)
    tau = torch.as_tensor(timepoints)
    St = torch.as_tensor(estimates)

    Gt = IPCW(futime, status)(tau, lag=0)
    GT = IPCW(futime, status)(futime, lag=1)

    out = torch.zeros_like(St)
    for i, (ti, s) in enumerate(zip(futime, status)):
        for j, t in enumerate(tau):
            for k in range(2):
                p = St[i][j][k]
                if ti <= t and s - 1 == k:
                    out[i, j, k] = (1 - p) ** 2 / GT[i]
                elif ti <= t and s != 0:
                    out[i, j, k] = p**2 / GT[i]
                elif ti > t:
                    out[i, j, k] = p**2 / Gt[j]

    return torch.mean(out, dim=0)


@given(
    tau=st.lists(st.floats(0, 24.3), min_size=1, max_size=100),
    est=st.floats(0, 1),
)
def test_brier_score_2(tau, est, survival_data_2):
    time, event = survival_data_2
    brier_score = BrierScore(survival_train=(time, event))
    phi = torch.full((len(time), len(tau), 2), est)

    # the discotime-packaged implementation
    brier_score_discotime = brier_score(
        estimates=phi,
        timepoints=tau,
        survival_test=(time, event),
    )

    assert brier_score_discotime.shape == phi.shape[1:]
    assert torch.all(
        (0 <= brier_score_discotime) & (brier_score_discotime <= 1)
    )

    # the stupid implementation relying on nested for-loops
    brier_score_naive = brier_score_nested(
        estimates=phi,
        timepoints=tau,
        survival_test=(time, event),
    )

    assert torch.all(brier_score_discotime == brier_score_naive)


@given(
    tau=st.lists(st.floats(0, 24.3), min_size=1, max_size=10),
    est=st.floats(0, 1),
)
def test_brier_score_3(tau, est, survival_data_2):
    """
    Guessing the incidence using the AalenJohansen estimator should be at least
    as good as any random constant guess.
    """
    time, event = survival_data_2
    brier_score = BrierScore(survival_train=(time, event))

    phi_test = torch.full((len(time), len(tau), 2), est)
    phi_null = AalenJohansen(time, event)(tau)
    phi_null = torch.as_tensor(phi_null).view(1, -1, 2).expand_as(phi_test)

    get_score = lambda x: brier_score(x, tau, (time, event))

    assert torch.all(get_score(phi_null).mean() <= get_score(phi_test).mean())


@pytest.mark.parametrize(
    ("cause", "tau", "expected"),
    [
        (1, [8.67], [0.192]),
        (1, [15.77, 1.77], [0.04, 0.231]),
        (1, [20.4, 2.57, 11.95, 12.85], [0.04, 0.231, 0.231, 0.249]),
        (1, [8.71], [0.192]),
        (1, [16.41, 20.18], [0.231, 0.247]),
        (2, [21.84, 20.91, 18.83], [0.235, 0.235, 0.235]),
        (2, [20.31, 12.49, 2.92, 0.56], [0, 0.076, 0.235, 0.235]),
        (2, [23.38, 18.46, 15.24], [0.235, 0.235, 0.235]),
    ],
)
def test_brier_score_null(tau, cause, expected, survival_data_2):
    """Expected values obtained using the `riskRegression` R package.

    ```{r}
    cfit <- CSC(Hist(t, e) ~ 1, cause = _, data = data)
    Score(list(cfit), formula=Hist(t, e) ~ 1, data=data, times=c(...), cause=_)
    ```
    """
    futime, status = survival_data_2
    tau = sorted(tau)

    phi_null = AalenJohansen(futime, status)(tau)
    phi_null = (
        torch.as_tensor(phi_null)
        .view(1, -1, 2)
        .expand((len(futime), len(tau), 2))
    )

    brier_score = BrierScore(survival_train=(futime, status))
    result = brier_score(phi_null, tau, (futime, status))[..., cause - 1]
    assert result == pytest.approx(expected, abs=1e-3)


@given(
    tau=st.lists(st.floats(0, 50), min_size=2, unique=True),
    est=st.floats(0, 1),
    integrate=st.booleans(),
)
@example(
    tau=[0.0, 3.503246160812043e-46],
    est=0.0,
    integrate=True,
)
def test_brier_score_scaled_1(tau, est, integrate, survival_data_2):
    tau = sorted(tau)
    time, event = survival_data_2
    brier_score_scaled = BrierScoreScaled(
        survival_train=(time, event), eval_grid=tau, integrate=integrate
    )
    phi = torch.full((len(time), len(tau), 2), est)
    bss = brier_score_scaled(phi, survival_data_2)

    expected_ndim = 1 if integrate else 2
    assert bss.ndim == expected_ndim

    expected_shape = (2,) if integrate else (len(tau), 2)
    assert bss.shape == expected_shape

    assert torch.all(bss <= 1)


def test_brier_score_scaled_2(default_mgus2_model):
    model = default_mgus2_model
    data = [*map(torch.as_tensor, model.datamodule.dset_fit[:])]

    tau = torch.linspace(10, 365, 50)
    bss = BrierScoreScaled((data[3], data[4]), tau, True)

    def training_loop():
        optim = model.configure_optimizers()

        for i in count():
            model.train()
            for batch_idx, batch in enumerate(
                model.datamodule.train_dataloader()
            ):
                optim.zero_grad()
                loss = model.training_step(batch, batch_idx)
                model.backward(loss)
                optim.step()

            model.eval()
            est = model._predict_estimates(data[0], tau)

            if i % 5 == 0:
                yield bss(est, (data[3], data[4]))

    assert all(
        torch.all(a <= (b + 0.01))
        for (a, b) in pairwise(islice(training_loop(), 20))
    )
